#include <glib-unix.h>
#include <gio/gio.h>
#include <stdbool.h>
#include "files-monior.h"


struct _DiyaFilesMonitor
{
    DiyaObject parent;
    GHashTable *watch_table;
};


G_DEFINE_FINAL_TYPE(DiyaFilesMonitor, diya_files_monitor, DIYA_TYPE_OBJECT)

static void diya_files_monitor_finalize(GObject *object)
{
    DiyaFilesMonitor *self = DIYA_FILES_MONITOR(object);
    g_debug("diya_files_monitor_finalize: %s", diya_object_to_string(object));
    g_hash_table_destroy(self->watch_table);
    G_OBJECT_CLASS(diya_files_monitor_parent_class)->finalize(object);
}

static void diya_files_monitor_dispose(GObject *object)
{
    // DiyaFilesMonitor *self = DIYA_FILES_MONITOR(object);
    g_debug("diya_files_monitor_dispose: %s", diya_object_to_string(object));
    G_OBJECT_CLASS(diya_files_monitor_parent_class)->dispose(object);
}

static void diya_files_monitor_init(DiyaFilesMonitor *self)
{
    self->watch_table = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_object_unref);
}

static const gchar *diya_files_monitor_to_string(DiyaObject *object)
{
    (void) object;
    return "Diya Files Monitor";
}

static void diya_files_monitor_class_init(DiyaFilesMonitorClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);

    gobject_class->finalize = diya_files_monitor_finalize;
    gobject_class->dispose = diya_files_monitor_dispose;
    base_class->to_string = diya_files_monitor_to_string;
}

DiyaFilesMonitor* diya_files_monitor_new()
{
    DiyaFilesMonitor* self = DIYA_FILES_MONITOR(g_object_new(DIYA_TYPE_FILES_MONITOR, NULL));
    return self;
}

gboolean diya_files_monitor_watch(DiyaFilesMonitor* self, const gchar* path, GCallback callback, gpointer userdata)
{
    GFileMonitor* fm = NULL;
    if(!g_hash_table_contains(self->watch_table, path))
    {
        GError *err = NULL;
        GFile* file = g_file_new_for_path(path);
        fm = g_file_monitor(file, G_FILE_MONITOR_WATCH_MOVES, NULL, &err);
        g_object_unref(file);
        if (err)
        {
            g_warning("Unable to watch file %s: %s", path, err->message);
            g_error_free(err);
            return false;
        }
        g_info("DiyaFilesMonitor: watch new file: %s", path);
        g_hash_table_insert(self->watch_table, (gpointer)g_strdup(path), fm);
    }
    else
    {
        g_info("File %s is already watched, register callback handle to the existing watcher", path);
        fm = g_hash_table_lookup(self->watch_table, path);
    }
    g_signal_connect(G_OBJECT(fm), "changed", callback, userdata);
    return true;
}

void diya_files_monitor_unwatch(DiyaFilesMonitor* self, const gchar* path)
{
    if(!g_hash_table_contains(self->watch_table, path))
    {
        g_hash_table_remove(self->watch_table, (gconstpointer)path);
    }
}