#include <wayland-client-protocol.h>
#include <gdk/wayland/gdkwayland.h>
#include <assert.h>
#include "launcher.h"
#include "widgets/dashboard-widget.h"
#include "widgets/taskbar-widget.h"
#include "foreign.h"
#include "wayland.h"

struct _DiyaLauncher
{
    DiyaShellObject parent;
    DiyaDashboardWidget* dashboard;
    DiyaTaskbarWidget* taskbar;
};
G_DEFINE_FINAL_TYPE(DiyaLauncher, diya_launcher, DIYA_TYPE_SHELL_OBJECT)

enum
{
    NO_PROP,
    PROP_DASHBOARD,
    PROP_TASKBAR,
    N_PROPERTIES
};

static GParamSpec *g_prop[N_PROPERTIES] = {0};

/*
static void show_windows(GtkWidget *widget,gpointer data)
{
    (void) widget;
    g_warning("Show all windows");
    DiyaSessionShell * shell = DIYA_SESSION_SHELL(data);
    assert(shell);
    GHashTable* windows = NULL;
    g_object_get(shell, DIYA_PROP_SESSION_WINDOWS, &windows, NULL);
    assert(windows);
    GHashTableIter iter;
    gpointer handle;
    DiyaForeignWindow* window;
    g_hash_table_iter_init (&iter, windows);
    while (g_hash_table_iter_next(&iter, (gpointer) &handle, (gpointer) &window))
    {
        g_warning("unset minimized %s", diya_object_to_string(window));
        diya_foreign_window_set_state(window, DIYA_WIN_STATE_MINIMIZE, false);
    }
}

*/

static void diya_launcher_dispose(GObject* object)
{
    (void) object;
    DiyaLauncher * self = DIYA_LAUNCHER(object);
    g_debug("diya_launcher_dispose: %s", diya_object_to_string(object));
    if(self->dashboard)
    {
        gtk_window_destroy(GTK_WINDOW(self->dashboard));
        self->dashboard = NULL;
    }
    if(self->taskbar)
    {
        gtk_window_destroy(GTK_WINDOW(self->taskbar));
        self->taskbar = NULL;
    }
    G_OBJECT_CLASS(diya_launcher_parent_class)->dispose(object);
}

static void diya_launcher_init(DiyaLauncher * self)
{
    self->taskbar = NULL;
    self->dashboard = NULL;
}

static const gchar* diya_launcher_to_string(DiyaObject* object)
{
    (void) object;
    return "DiyaLauncher";
}


static void diya_launcher_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec)
{
    DiyaLauncher * self = DIYA_LAUNCHER(object);

    switch (property_id)
    {
    case PROP_DASHBOARD:
        self->dashboard = g_value_get_pointer(value);
        break;
    case PROP_TASKBAR:
        self->taskbar = g_value_get_pointer(value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_launcher_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec)
{
    DiyaLauncher * self = DIYA_LAUNCHER(object);
    switch (property_id)
    {
    case PROP_DASHBOARD:
        g_value_set_pointer(value, self->dashboard);
        break;
    case PROP_TASKBAR:
        g_value_set_pointer(value, self->taskbar);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void diya_launcher_class_init(DiyaLauncherClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(class);
    DiyaObjectClass *base_class = DIYA_OBJECT_CLASS(class);

    gobject_class->dispose = diya_launcher_dispose;
    gobject_class->set_property = diya_launcher_set_property;
    gobject_class->get_property = diya_launcher_get_property;
    base_class->to_string = diya_launcher_to_string;

    g_prop[PROP_DASHBOARD] = g_param_spec_pointer(DIYA_PROP_LAUNCHER_DASHBOARD, NULL, "Launcher dashboard", G_PARAM_READWRITE|G_PARAM_CONSTRUCT_ONLY);
     g_prop[PROP_TASKBAR] = g_param_spec_pointer(DIYA_PROP_LAUNCHER_TASKBAR, NULL, "Launcher taskbar", G_PARAM_READWRITE|G_PARAM_CONSTRUCT_ONLY);
    g_object_class_install_properties (gobject_class, N_PROPERTIES, g_prop);
}

void diya_session_shell_launcher_init(DiyaSessionShell * shell)
{
    DiyaDashboardWidget* dashboard = DIYA_DASHBOARD_WIDGET(diya_shell_window_new(DIYA_TYPE_DASHBOARD_WIDGET, shell));
    assert(dashboard);
    DiyaTaskbarWidget* taskbar = DIYA_TASKBAR_WIDGET(diya_shell_window_new(DIYA_TYPE_TASKBAR_WIDGET, shell));
    assert(taskbar);
    //gtk_window_set_transient_for(GTK_WINDOW(dashboard), GTK_WINDOW(taskbar));
    DiyaLauncher* self = DIYA_LAUNCHER(g_object_new (DIYA_TYPE_LAUNCHER, "dashboard", dashboard, "taskbar",taskbar, NULL));
    assert(self);
    g_object_bind_property(taskbar, DIYA_PROP_TASKBAR_SHOW_DASHBOARD, dashboard, DIYA_PROP_DASHBOARD_ACTIVE, G_BINDING_BIDIRECTIONAL);
    g_object_set(shell, "launchpad", self, NULL);
    //g_signal_connect (GTK_WINDOW(dashboard), "destroy", G_CALLBACK (on_launcher_destroy), NULL);

    gtk_window_present (GTK_WINDOW (taskbar));

    GtkIconTheme* theme = gtk_icon_theme_get_for_display(gdk_display_get_default());
    char**icons = gtk_icon_theme_get_icon_names (theme);
    g_warning("List all icons");
    for(int i = 0; icons[i] != NULL; i++)
    {
        g_debug("ICON: %s", icons[i]);
    }
}

void diya_launcher_show_dashboard(DiyaLauncher* self, gboolean value)
{
    g_object_set(self->dashboard, DIYA_PROP_DASHBOARD_ACTIVE, value, NULL);
}


void diya_launcher_toggle_dashboard(DiyaLauncher* self)
{
    diya_launcher_show_dashboard(self, !diya_dashboard_is_active(self->dashboard));
}